use glam::{DMat3, DVec3, Vec3};
use noise::NoiseFn;
use rand::prelude::*;
use sparse::Sparse;
use terrain::{
    anchor::Anchor,
    generator::Generator,
    icosphere::Icosphere,
    mesh::Mesh,
    terrain::Terrain,
    utility::rgb,
};

fn get_subdivisions(radius_km:f64) -> usize
// Determines the number of subdivisions required for an icosphere,
// given a world radius, such that the distance between triangle
// centers is approximately 1-2 kilometers (assuming units of meters).
//
{
    let edge_length = 4. * radius_km / (10. + (2. * f64::sqrt(5.))).sqrt();
    let cell_radius = edge_length / f64::sqrt(3.);
    
    let mut count = 0;
    let mut r = cell_radius;
    while r > 2. { r /= 2.; count += 1; }
    count
}

fn get_neighbors(ico:&Icosphere, origin:usize, distance:usize) -> Vec<(usize, usize)>
// (Id, Distance)
// Gets a unique list of icosphere cells up to a specified distance from a origin cell.
//
{
    let mut search = Vec::<(usize, usize)>::new();
    let mut found = Sparse::<()>::new();
    found.set(origin as isize, ());

    let mut queue = std::collections::VecDeque::<(usize, usize)>::new();
    queue.push_back((origin, 0));

    while let Some((id, depth)) = queue.pop_front() {
        search.push((id, depth));
        
        if let Some(cell) = ico.cells.get(id) {
            if depth < distance {
                for nid in &cell.neighbors {
                    if found.get(*nid as isize).is_none() {
                        found.set(*nid as isize, ());
                        queue.push_back((*nid, depth + 1));
                    }
                }
            }
        }
    }

    search
}

fn main() {
    let radius_km = 10.;
    let terrain_radius = 6;
    let max_terrain_detail = 14;

    let mut ico = Icosphere::new();
    let mut anchors = Sparse::<Anchor>::new();
    let subdivisions = get_subdivisions(radius_km);
    for _ in 0..subdivisions { ico.subdivide(); }
    println!("Subdivisions: {}", subdivisions);

    let world_radius = radius_km * 1000.;
    let world_relief = world_radius / 150.;  // typically 100:1 - 300:1

    println!("radius: {} / relief: {}", world_radius, world_relief);

    let mut rng = rand::thread_rng();
    let perlin = noise::Perlin::new(12);

    // Generate an anchor for each cell in the isosphere.
    //
    let cells_list = ico.cells.list();
    let mut anchor_counter = 0;
    let anchor_counter_total = cells_list.len();
    for id in cells_list {
        if anchor_counter % 10000 == 0 {
            println!("Generating anchor block #{}/{}", anchor_counter, anchor_counter_total);
        } anchor_counter += 1;

        let cell = ico.cells.get_mut(id).unwrap().clone();

        // Calculate cell space and transformations.
        let vertices = [
            *ico.vertices.get(cell.vertices[0]).unwrap(),
            *ico.vertices.get(cell.vertices[1]).unwrap(),
            *ico.vertices.get(cell.vertices[2]).unwrap(),
        ];
        
        let centroid = (vertices[0] + vertices[1] + vertices[2]) / 3.;
        
        let basis_i = (vertices[1] - vertices[0]).normalize();
        let basis_j = basis_i.cross(vertices[2] - vertices[0]).normalize();
        let basis_k = basis_j.cross(basis_i).normalize();

        let basis_inverse = DMat3::from_cols(basis_i, basis_j, basis_k);
        let basis = basis_inverse.inverse();

        // Define anchor properties.
        //
        anchors.set(id as isize, {
            let mut anchor = Anchor::new();

            anchor.centroid = centroid;
            anchor.position = centroid.normalize();
            anchor.basis = basis;
            anchor.ibasis = basis_inverse;

            let mut elevation_frequency = 1. / 4.;
            for _ in 0..4 {
                let frequency = world_radius * elevation_frequency;
                anchor.geography.elevation += (world_relief * elevation_frequency) * perlin.get([frequency * anchor.position.x, frequency * anchor.position.y, frequency * anchor.position.z]);
                elevation_frequency /= 10.;
            }

            let octaves = (rng.gen::<f64>() * anchor.geography.octaves.len() as f64) as usize + 1;
            let mut octave_modifier = 500.;
            for oct in 0..octaves {
                let wavelength = (octave_modifier * world_relief * rng.gen::<f64>()) + 10.;
                anchor.geography.octaves[oct].frequency = world_radius / wavelength;
                anchor.geography.octaves[oct].amplitude = f64::sqrt(0.25 * rng.gen::<f64>() * wavelength);
                anchor.geography.octaves[oct].offset = world_radius * DVec3::new(
                    (rng.gen::<f64>() * 2.) - 1.,
                    (rng.gen::<f64>() * 2.) - 1.,
                    (rng.gen::<f64>() * 2.) - 1.,
                );
                octave_modifier /= (rng.gen::<f64>() * 100.) + 2.;
            }

            anchor
        });

        ico.cells.update(id, cell).ok();
    }


    // Set anchor neighbor properties.
    for id in ico.cells.list() {
        let cell = ico.cells.get(id).unwrap();
        let mut anchor = anchors.get(id as isize).unwrap().clone();

        for n in 0..cell.neighbors.len() {
            let neighbor_id = cell.neighbors[n];
            let neighbor_anchor = anchors.get(neighbor_id as isize).unwrap();

            anchor.neighbors[n].distance = anchor.position.angle_between(neighbor_anchor.position) * world_radius;
            anchor.neighbors[n].position = (anchor.basis * (neighbor_anchor.position - anchor.position)) * world_radius;
            anchor.neighbors[n].transform = anchor.ibasis * neighbor_anchor.basis;
        }

        anchors.set(id as isize, anchor);
    }


    // Generate terrain.
    let mut terrain_mesh: Mesh = Mesh::new();

    let mut mesh_count = 0;
    let origin_id = 1;

    let origin_anchor = anchors.get(origin_id as isize).unwrap().clone();
    let world_origin = DVec3::NEG_Y * world_radius;

    // Generate zone meshes.
    let tiles = get_neighbors(&ico, 1, terrain_radius);
    let total_tiles = tiles.len();
    for (tile_id, depth) in &tiles {

        let mut gen = Generator::new();

        mesh_count += 1;
        println!("Generating {} [{} / {}]", tile_id, mesh_count, total_tiles);

        let terrain = if let Some(zone) = ico.cells.get(*tile_id) {
            // Prepare generator with parameters of neighboring tiles.
            let nearby = get_neighbors(&ico, *tile_id, 3);
            for vindex in 0..zone.vertices.len() {
                let mut influence_count = 0;

                let extent = zone.vertices[vindex];
                gen.sources[vindex] = terrain::generator::Source::new();
                gen.sources[vindex].position = *ico.vertices.get(extent).unwrap();

                // Average geography of nearby anchors for each tile vertex.
                for n in 0..nearby.len() {
                    let (nearby_id, _depth) = nearby[n];

                    if let Some(tile) = ico.cells.get(nearby_id) {
                        let tile_anchor = anchors.get(nearby_id as isize).unwrap();
                        for nvid in tile.vertices {
                            if nvid == extent {
                                gen.sources[vindex].elevation += tile_anchor.geography.elevation;
                                gen.sources[vindex].octaves.append(&mut tile_anchor.geography.octaves.to_vec());
                                influence_count += 1;
                                break;
                            }
                        }
                    }
                }

                gen.sources[vindex].elevation /= influence_count as f64;
            }

            let mut terrain = Terrain::triangle((gen.sources[0].position, gen.sources[1].position, gen.sources[2].position), gen);
            terrain.config.origin = origin_anchor.position;
            terrain.config.scale = world_radius;

            for _ in 0..(4 - (*depth).min(3)) { terrain.subdivide(); }
            terrain.tessellate(max_terrain_detail);
            terrain
        } else { break; };

        anchors.get_mut(*tile_id as isize).unwrap().terrain = terrain;
    }

    // Modify anchor terrain to match neighbor edge subdivisons.
    let mut apply_count = 1;
    while apply_count > 0 {
        apply_count = 0;

        let mut tile_count = 0;
        for (tile_id, _depth) in &tiles {
            let tile = ico.cells.get(*tile_id).unwrap().clone();
            let mut cterrain = anchors.get(*tile_id as isize).unwrap().terrain.clone();

            tile_count += 1;
            println!("Stitching {} [{} / {}]", *tile_id, tile_count, tiles.len());

            // Apply stitching to each side of the triangle mesh.
            for tile_side_index in 0..tile.neighbors.len() {
                let neighbor_id = tile.neighbors[tile_side_index];
                let neighbor_tile = ico.cells.get(neighbor_id).unwrap();
                let nterrain = &mut anchors.get_mut(neighbor_id as isize).unwrap().terrain;

                if nterrain.extent == [0; 3] {
                    // Ignore uninitialized (out of scope) terrain.
                    continue;
                }

                // Find indices of shared vertices.
                let mut shared_vertex_indices = [0, 0];
                let mut nshared_vertex_indices = [0, 0];
                let mut shared_vertices = [0, 0];
                let mut nshared_vertices = [0, 0];
                let mut shared_index = 0;
                for vertex_index in 0..tile.vertices.len() {
                    for nvertex_index in 0..neighbor_tile.vertices.len() {
                        if shared_index < 2 && tile.vertices[vertex_index] == neighbor_tile.vertices[nvertex_index] {
                            shared_vertex_indices[shared_index] = vertex_index;
                            nshared_vertex_indices[shared_index] = nvertex_index;

                            shared_vertices[shared_index] = tile.vertices[vertex_index];
                            nshared_vertices[shared_index] = neighbor_tile.vertices[nvertex_index];

                            shared_index += 1;
                        }
                    }
                }

                let neighbor_direction = 3 - (nshared_vertex_indices[0] + nshared_vertex_indices[1]);

                /*println!("shared c({} : {}={}/{}, {}={}/{}) n({} : {}={}/{}, {}={}/{})",
                    tile_side_index,
                    shared_vertex_indices[0], shared_vertices[0], cterrain.extent[shared_vertex_indices[0]],
                    shared_vertex_indices[1], shared_vertices[1], cterrain.extent[shared_vertex_indices[1]],
                    neighbor_direction,
                    nshared_vertex_indices[0], nshared_vertices[0], nterrain.extent[nshared_vertex_indices[0]],
                    nshared_vertex_indices[1], nshared_vertices[1], nterrain.extent[nshared_vertex_indices[1]],
                );*/

                // Get ids of corner cells.
                let mut current_face_id = {
                    let mut find = 0;
                    for face_id in cterrain.cells.list() { 
                        let face = cterrain.cells.get(face_id).unwrap();
                        let search = cterrain.extent[shared_vertex_indices[0]];
                        if face.vertices[0] == search || face.vertices[1] == search || face.vertices[2] == search {
                            find = face_id;
                            break;
                        }
                    }
                    find
                };
                let mut neighbor_face_id = {
                    let mut find = 0;
                    for face_id in nterrain.cells.list() {
                        let face = nterrain.cells.get(face_id).unwrap();
                        let search = nterrain.extent[nshared_vertex_indices[0]];
                        if face.vertices[0] == search || face.vertices[1] == search || face.vertices[2] == search {
                            find = face_id;
                            break;
                        }
                    }
                    find
                };

                if current_face_id == 0 || neighbor_face_id == 0 {
                    //println!("! Failed to find faces: {}, {}", current_face_id, neighbor_face_id);
                    continue;
                }

                let current_direction_down = shared_vertex_indices[0];
                let current_direction_edge = tile_side_index;
                let current_direction_away = 3 - (current_direction_down + current_direction_edge);
                let neighbor_direction_down = nshared_vertex_indices[0];
                let neighbor_direction_edge = neighbor_direction;
                let neighbor_direction_away = 3 - (neighbor_direction_down + neighbor_direction_edge);

                /*println!("  DIR D0 {} E0 {} A0 {} D1 {} E1 {} A1 {}",
                    current_direction_down,
                    current_direction_edge,
                    current_direction_away,
                    neighbor_direction_down,
                    neighbor_direction_edge,
                    neighbor_direction_away,
                );*/

                let mut end_current = false;
                let mut end_neighbor = false;

                // Check if each edge face requires subdivision.
                //
                loop {
                    let tile_cell = cterrain.cells.get(current_face_id).unwrap().clone();
                    let neighbor_cell = nterrain.cells.get(neighbor_face_id).unwrap().clone();

                    /*println!("instep t {} (d {} r {}) n {} (d {} r {})",
                        current_face_id, tile_cell.depth, tile_cell.direction,
                        neighbor_face_id, neighbor_cell.depth, neighbor_cell.direction,
                    );*/

                    // Work is done once both cells have no neighbor in donward directions.
                    if end_current && end_neighbor {
                        //println!("done");
                        break;
                    }

                    //println!("step t({}) = {} n({}) = {}", current_face_id, tile_cell.depth, neighbor_face_id, neighbor_cell.depth);

                    if tile_cell.depth & !1 > neighbor_cell.depth & !1 {
                        // Subdivide neighbor tile's terrain and traverse backward.
                        nterrain.subdivide_cell(neighbor_face_id);

                        let new_face = nterrain.cells.get(neighbor_face_id).unwrap();
                        if new_face.sibling != 3 {
                            neighbor_face_id = new_face.neighbors[new_face.sibling as usize];
                        }

                        neighbor_face_id = nterrain.cells.get(neighbor_face_id).unwrap().neighbors[neighbor_direction_down];
                        apply_count += 1;

                    } else if tile_cell.depth & !1 < neighbor_cell.depth & !1 {
                        // Subdivide current tile's terrain and traverse backward.
                        cterrain.subdivide_cell(current_face_id);

                        let new_face = cterrain.cells.get(current_face_id).unwrap();
                        if new_face.sibling != 3 {
                            current_face_id = new_face.neighbors[new_face.sibling as usize];
                        }
                        current_face_id = cterrain.cells.get(current_face_id).unwrap().neighbors[current_direction_down];
                        apply_count += 1;

                    } else {
                        // Traverse to next tile on edge.
                        let mut move_last = current_face_id;
                        let mut move_direction = current_direction_down;
                        let mut move_count = 0;
                        while {
                            let face = cterrain.cells.get(current_face_id).unwrap();
                            
                            let neighbor_id = face.neighbors[move_direction];
                            /*println!("mv ({}) current {} / {} (d {} s {} r {}) -> {}",
                                move_direction,
                                move_last, neighbor_face_id,
                                face.depth, face.sibling, face.direction,
                                neighbor_id
                            );*/
                            if neighbor_id != 0 {
                                move_count += 1;

                                if move_last != neighbor_id {
                                    move_last = current_face_id;
                                    current_face_id = neighbor_id;
                                } else {
                                    if move_direction != current_direction_away {
                                        move_direction = current_direction_away;
                                    } else {
                                        move_direction = current_direction_edge;
                                    }
                                }
                                true
                            } else {
                                if move_count == 0 {
                                    end_current = true;
                                }
                                false
                            }
                        } { }

                        // Traverse neighbor tile to next edge face.
                        move_last = neighbor_face_id;
                        move_direction = neighbor_direction_down;
                        move_count = 0;
                        while {
                            let face = nterrain.cells.get(neighbor_face_id).unwrap();
                            
                            let neighbor_id = face.neighbors[move_direction];
                            /*println!("mv ({}) neighbor {} / {} (d {} s {} r {}) -> {}",
                                move_direction,
                                move_last, neighbor_face_id,
                                face.depth, face.sibling, face.direction,
                                neighbor_id
                            );*/
                            if neighbor_id != 0 {
                                move_count += 1;

                                if move_last != neighbor_id {
                                    move_last = neighbor_face_id;
                                    neighbor_face_id = neighbor_id;
                                } else {
                                    if move_direction != neighbor_direction_away {
                                        move_direction = neighbor_direction_away;
                                    } else {
                                        move_direction = neighbor_direction_edge;
                                    }
                                }
                                true
                            } else {
                                if move_count == 0 {
                                    end_neighbor = true;
                                }
                                false
                            }
                        } { }
                        
                    }
                }
            }

            anchors.get_mut(*tile_id as isize).unwrap().terrain = cterrain;
        }
    }

    // Write anchors to mesh.
    for (tile_id, _depth) in &tiles {
        let anchor = anchors.get_mut(*tile_id as isize).unwrap();

        for vid in anchor.terrain.vertices.list() {
            if let Some(vertex) = anchor.terrain.vertices.get_mut(vid) {
                let norm = vertex.position.normalize() * (world_radius + vertex.height);
                vertex.position = (origin_anchor.basis * norm) + world_origin;

                if vertex.height > 0. {
                    vertex.color = rgb(192, 128, 32);
                } else {
                    vertex.color = rgb(32, 96, 192);
                }
                vertex.height = 0.;
            }
        }

        terrain_mesh = terrain_mesh.merge(&anchor.terrain.to_mesh(), Vec3::ZERO);
    }

    terrain::export::export_obj("test.obj", &terrain_mesh, 0.1).ok();
}
