#![allow(dead_code)]

use std::f64::consts::PI;
use glam::{IVec3, DVec3, Vec3};
use pool::Pool;
use sparse::Sparse;

const DEBUG_PRINTS :bool = false;
const MESH_SCALE :f32 = 1.;

const FLIPPED :u32 = 1 << 31;
const NFLIPPED :u32 = !FLIPPED;

#[derive(Clone, Copy)]
pub struct Cell {
    pub data:usize,
    depth:u32,
    pub neighbors:[usize;3],
    pub vertices:[usize;3],
}
impl Cell {
    pub fn new() -> Self
    {
        Self {
            data:0,
            depth:0,
            neighbors:[0; 3],
            vertices:[0; 3],
        }
    }
}

pub struct Icosphere {
    pub vertices:Pool<DVec3>,
    pub cells:Pool<Cell>,
}
impl Icosphere {
    pub fn new() -> Self
    {
        Self {
            vertices:Pool::<DVec3>::new(),
            cells:Pool::<Cell>::new(),
        }.init()
    }

    fn init(mut self) -> Self
    {
        let radius = f64::sqrt(1.25);

        let k = (2. / 5.) * PI;

        // Generate initial structure
        self.add_vertex(DVec3::new(0., radius, 0.));
        for i in 0..5 {
            let theta = k * i as f64;
            self.add_vertex(DVec3::new(
                f64::cos(theta), 
                0.5,
                f64::sin(theta),
            ));
        }
        for i in 0..5 {
            let theta = k * (i as f64 + 0.5);
            self.add_vertex(DVec3::new(
                f64::cos(theta), 
                -0.5,
                f64::sin(theta),
            ));
        }
        self.add_vertex(DVec3::new(0., -radius, 0.));

        // Top row cells
        for i in 0..5 {
            self.cells.add({
                let mut cell = Cell::new();

                cell.vertices = [
                    2 + ((i + 1) % 5),
                    2 + i,
                    1,
                ];
                cell.neighbors = [
                    ((i + 4) % 5) + 1,
                    ((i + 1) % 5) + 1,
                    6 + (i * 2),
                ];

                cell
            });
        }

        // Center row cells
        for i in 0..10 {
            self.cells.add({
                let mut cell = Cell::new();

                // Top facing cells
                if i & 1 == 0 {
                    cell.depth |= FLIPPED;

                    cell.vertices = [
                        2 + ((i / 2) % 5),
                        2 + (((i / 2) + 1) % 5),
                        7 + ((i / 2) % 5),
                    ];
                    cell.neighbors = [
                        6 + ((i + 1) % 10),
                        6 + ((i + 9) % 10),
                        1 + (i / 2),
                    ];
                }

                // Bottom facing cells
                else {
                    cell.vertices = [
                        7 + (((i / 2) + 1) % 5),
                        7 + ((i / 2) % 5),
                        2 + (((i / 2) + 1) % 5),
                    ];
                    cell.neighbors = [
                        6 + ((i + 9) % 10),
                        6 + ((i + 1) % 10),
                        16 + (i / 2),
                    ];
                }

                cell
            });
        }

        // Bottom row cells
        for i in 0..5 {
            self.cells.add({
                let mut cell = Cell::new();

                cell.depth |= FLIPPED;

                cell.vertices = [
                    7 + i,
                    7 + ((i + 1) % 5),
                    12,
                ];
                cell.neighbors = [
                    16 + ((i + 1) % 5),
                    16 + ((i + 4) % 5),
                    7 + (i * 2),
                ];

                cell
            });
        }
        self
    }
    

    pub fn subdivide(&mut self)
    {
        let list = self.cells.list();
        for id in list {
            self.node_subdivide(id, 0);
        }
    }

    fn node_subdivide(&mut self, id:usize, _caller:usize)
    {
        let cell = if let Some(data) = self.cells.get(id) {
            data.clone()
        } else { return; };

        //println!("caller {}", caller);
        self.print_cell("BEGDIV", id, &cell);

        let mut updates = Vec::<usize>::new();
        for n in 0..cell.neighbors.len() {
            let neighbor_id = cell.neighbors[n];
            if let Some(neighbor) = self.cells.get(neighbor_id) {
                if neighbor.neighbors.contains(&id) {
                    updates.push(neighbor_id);
                }
                else {
                    let nd_offset_0 = (n + 2) % 3;
                    let nd_offset_1 = (n + 1) % 3;

                    updates.push(neighbor.neighbors[nd_offset_0]);
                    updates.push(neighbor.neighbors[nd_offset_1]);
                }
            }
        }

        // Get vector values of outer vertices.
        let vertex_values = [
            *self.vertices.get(cell.vertices[0]).unwrap(),
            *self.vertices.get(cell.vertices[1]).unwrap(),
            *self.vertices.get(cell.vertices[2]).unwrap(),
        ];

        // Get midpoints from adjacent nodes.
        let mut midpoints = [0; 3];
        for n in 0..cell.neighbors.len() {
            let neighbor_id = cell.neighbors[n];
            if neighbor_id != 0 {
                let neighbor = self.cells.get(neighbor_id).unwrap().clone();
                
                if neighbor.depth & NFLIPPED > cell.depth & NFLIPPED {

                    if neighbor.depth & FLIPPED == cell.depth & FLIPPED {
                        midpoints[n] = neighbor.vertices[n];
                    } else {
                        midpoints[n] = neighbor.vertices[(n == 0) as usize];
                    }
                }
            }
        }

        // Fill missing midpoint vertices.
        let subvertices = [
            if midpoints[0] != 0 { midpoints[0] } else { self.add_vertex(((vertex_values[2] - vertex_values[1]) / 2.) + vertex_values[1]) },
            if midpoints[1] != 0 { midpoints[1] } else { self.add_vertex(((vertex_values[0] - vertex_values[2]) / 2.) + vertex_values[2]) },
            if midpoints[2] != 0 { midpoints[2] } else { self.add_vertex(((vertex_values[1] - vertex_values[0]) / 2.) + vertex_values[0]) },
        ];


        // Prepare child cells.
        let mut children = [
            Cell::new(),
            Cell::new(),
            Cell::new(),
            Cell::new(),
        ];
        let child_ids = [
            self.cells.add(Cell::new()),
            self.cells.add(Cell::new()),
            self.cells.add(Cell::new()),
            id,
        ];

        for c in 0..children.len() {
            children[c].depth = cell.depth + 1;
        }
        for c in 0..3 {
            children[3].neighbors[c] = child_ids[c];
            children[c].neighbors[c] = id;
        }
        children[3].depth ^= FLIPPED;

        // Set new cell vertices.
        children[0].vertices = [ cell.vertices[0],   subvertices[2],   subvertices[1] ];
        children[1].vertices = [   subvertices[2], cell.vertices[1],   subvertices[0] ];
        children[2].vertices = [   subvertices[1],   subvertices[0], cell.vertices[2] ];
        children[3].vertices = [   subvertices[0],   subvertices[1],   subvertices[2] ];

        // Trade neighbor links with neighboring cells.
        // If neighbor is at new depth, link child nodes.
        for n in 0..cell.neighbors.len() {
            let neighbor_id = cell.neighbors[n];

            let nd_offset_0 = (n + 2) % 3;
            let nd_offset_1 = (n + 1) % 3;

            if neighbor_id != 0 {
                let neighbor = self.cells.get(neighbor_id).unwrap().clone();

                if neighbor.depth & NFLIPPED > cell.depth & NFLIPPED {
                    let (c0, c1, nc0, nc1, nd) = if cell.depth & FLIPPED == neighbor.depth & FLIPPED {(
                        nd_offset_1,
                        nd_offset_0,
                        nd_offset_0,
                        nd_offset_1,
                        n,
                    )} else {(
                        2,
                        (n == 0) as usize,
                        2,
                        n,
                        (n == 0) as usize,
                    )};

                    let n0 = neighbor.neighbors[nc0];
                    let n1 = neighbor.neighbors[nc1];

                    //println!("{} [{}] = {}, {} [{}] = {}", n0, n, child_ids[nd_offset_0], n1, n, child_ids[nd_offset_1]);

                    children[c0].neighbors[n] = n0;
                    children[c1].neighbors[n] = n1;

                    if let Some(neighbor) = self.cells.get_mut(n0) {
                        neighbor.neighbors[nd] = child_ids[c0];
                    }
                    if let Some(neighbor) = self.cells.get_mut(n1) {
                        neighbor.neighbors[nd] = child_ids[c1];
                    }
                }
            }
        }

        // Update cells.
        for i in 0..children.len() {
            self.print_cell("SUBDIV", child_ids[i], &children[i]);

            let mut child = Cell::new();
            std::mem::swap(&mut children[i], &mut child);
            self.cells.update(child_ids[i], child).ok();
        }
    }

    fn add_vertex(&mut self, vertex:DVec3) -> usize
    {
        self.vertices.add(vertex.normalize())
    }

    fn print_cell(&self, text:&str, id:usize, cell:&Cell)
    {
        if DEBUG_PRINTS {
            println!("{:8} {:3} ( {}, {:2} ) [{:3} {:3} {:3}] [{:3} {:3} {:3}]",
                text,
                id,
                (cell.depth & FLIPPED != 0) as u8,
                cell.depth & NFLIPPED,
                cell.vertices[0],
                cell.vertices[1],
                cell.vertices[2],
                cell.neighbors[0],
                cell.neighbors[1],
                cell.neighbors[2],
            );
        }
    }

    pub fn to_mesh(&self) -> crate::mesh::Mesh
    {
        let mut mesh = crate::mesh::Mesh::new();
        let mut conversion = Sparse::<usize>::new();

        // Add vertices
        for index in self.vertices.list() {
            if let Some(vertex) = self.vertices.get(index) {
                conversion.set(index as isize, mesh.vertices.len());
                mesh.vertices.push(crate::mesh::Vertex::new(
                    vertex.as_vec3() * MESH_SCALE,
                    Vec3::new(1., 1., 1.),
                ));
            }
        }

        // Add faces
        for index in self.cells.list() {
            if let Some(cell) = self.cells.get(index) {
                self.print_cell("FINAL", index, &cell);

                let a = *conversion.get(cell.vertices[0] as isize).unwrap() as i32;
                let b = *conversion.get(cell.vertices[1] as isize).unwrap() as i32;
                let c = *conversion.get(cell.vertices[2] as isize).unwrap() as i32;
                
                mesh.faces.push(IVec3::new(a, b, c));
            }
        }

        mesh
    }
}
