use glam::DVec3;

mod orbit; pub use orbit::Orbit;

#[derive(Clone, Copy)]
struct Edge {
    pub id:usize,
    pub orbit:Orbit,
}
impl Edge {
    pub fn new(id:usize, orbit:Orbit) -> Self
    {
        Self {
            id:id,
            orbit:orbit,
        }
    }
}

#[derive(Clone, Copy)]
enum NodeData {
    Single(usize),
    Binary(Edge, Edge),
}

#[derive(Clone)]
struct Node {
    pub data:NodeData,
    pub mass:f64,
    pub children:Vec<Edge>,
}

#[derive(Clone)]
pub struct StarSystem {
    nodes:Vec<Node>,
}
impl StarSystem {
    pub fn singular(id:usize, mass:f64) -> Self
    {
        Self {
            nodes:vec![Node {
                data:NodeData::Single(id),
                mass:mass,
                children:vec![],
            }],
        }
    }

    pub fn binary(a:Self, b:Self, distance:f64, orbit:Orbit) -> Self
    {
        let offset_a = 1;
        let offset_b = a.nodes.len() + 1;

        let mass_a = a.nodes[0].mass;
        let mass_b = b.nodes[0].mass;
        let total_mass = mass_a + mass_b;

        let mut orbit_a = orbit;
        orbit_a.periapsis = mass_b * distance / total_mass;

        let mut orbit_b = orbit;
        orbit_b.periapsis = mass_a * distance / total_mass;

        Self {
            nodes:[
                vec![
                    Node {
                        data:NodeData::Binary(
                            Edge::new(offset_a, orbit_a),
                            Edge::new(offset_b, orbit_b),
                        ),
                        mass:total_mass,
                        children:vec![],
                    }
                ],
                a.nodes,
                b.nodes,
            ].concat(),
        }
    }

    pub fn append(mut self, mut system:Self, orbit:Orbit) -> Self
    {
        let id = self.nodes.len();
        println!("append at {}", id);

        for node in &mut system.nodes {
            for child in &mut node.children {
                child.id += id;
            }
        }

        self.nodes[0].children.push(Edge::new(id, orbit));
        self.nodes.append(&mut system.nodes.clone());
        self
    }

    pub fn resolve(&self, time:f64) -> Vec<(usize, DVec3)>
    // Returns the absolute-space positions of all bodies.
    //
    {
        let mut result = Vec::<(usize, DVec3)>::new();
        let mut stack = Vec::<(usize, DVec3)>::new();
        stack.push((0, DVec3::ZERO));

        while let Some(params) = stack.pop() {
            let (id, position) = params;
            println!("resolving {} at {}", id, position);

            match &self.nodes[id].data {
                NodeData::Single(eid) => {
                    result.push((*eid, position));
                }
                NodeData::Binary(edge1, edge2) => {
                    stack.push((edge1.id, position + edge1.orbit.position(time)));
                    stack.push((edge2.id, position - edge2.orbit.position(time)));
                }
            }

            for edge in &self.nodes[id].children {
                stack.push((edge.id, position + edge.orbit.position(time)));
            }
        }

        result
    }
}
