use std::f64::consts::PI;
use glam::{DVec3, DQuat};

#[derive(Clone, Copy)]
pub struct Orbit {
    pub orientation:DQuat,  // orientation of the orbital system
    pub periapsis:f64,      // focal distance at periapsis
    pub eccentricity:f64,   // eccentricity of orbit
    pub period:f64,         // period of orbit
    pub basis:f64,          // starting position in multiples of period
}
impl Orbit {
    pub fn new() -> Self
    {
        Self {
            orientation:DQuat::IDENTITY,
            periapsis:1.,
            eccentricity:0.,
            period:1.,
            basis:0.,
        }
    }

    pub fn with(orientation:DQuat, periapsis:f64, eccentricity:f64, period:f64, basis:f64) -> Self
    {
        Self {
            orientation:orientation,
            periapsis:periapsis,
            eccentricity:eccentricity,
            period:period,
            basis:basis,
        }
    }

    pub fn orientation(mut self, orientation:DQuat) -> Self
    {
        self.orientation = orientation;
        self
    }

    pub fn periapsis(mut self, periapsis:f64) -> Self
    {
        self.periapsis = periapsis;
        self
    }

    pub fn eccentricity(mut self, eccentricity:f64) -> Self
    {
        self.eccentricity = eccentricity;
        self
    }

    pub fn period(mut self, period:f64) -> Self
    {
        self.period = period;
        self
    }

    pub fn basis(mut self, basis:f64) -> Self
    {
        self.basis = basis;
        self
    }

    pub fn position(&self, time:f64) -> DVec3
    {
        let semimajor_axis = self.periapsis / (1. - self.eccentricity);
        let time = time + (self.basis * self.period);

        let n = 2. * PI / self.period;
        let mean_anomaly = n * time;

        let mut eccentric_anomaly = mean_anomaly + (self.eccentricity * ((mean_anomaly < PI) as i32 as f64 - 0.5));
        {
            let threshold = 1e-6;
            let mut delta: f64 = 1.;
            while delta.abs() > threshold {
                delta = (eccentric_anomaly - (self.eccentricity * eccentric_anomaly.sin()) - mean_anomaly) / (1. - self.eccentricity * eccentric_anomaly.cos());
                eccentric_anomaly = eccentric_anomaly - delta;
            }
        }

        let theta = 2. * ((1. + self.eccentricity).sqrt() * (eccentric_anomaly / 2.).tan()).atan2((1. - self.eccentricity).sqrt());
        let radius = semimajor_axis * (1. - (self.eccentricity * eccentric_anomaly.cos()));

        self.orientation * DVec3::new(
            radius * theta.cos(),
            0.,
            -radius * theta.sin(),
        )
    }
}
