#![allow(dead_code)]

use std::f64::consts::PI;
use glam::{IVec3, DVec3, Vec3};
use pool::Pool;
use sparse::Sparse;

use crate::generator::Generator;

const DEBUG_PRINTS :bool = false;

const DEGREES :f64 = PI / 180.;

const DELTA_THRESHOLD :f64 = (1. / 16.) * DEGREES;
const RADIAL_FALLOFF :f64 = 1. / 6.;

#[derive(Clone, Copy)]
enum Action {
    Check(/*Id*/usize, /*Caller*/usize),
    Subdivide(/*Id*/usize, /*Caller*/usize),
    Split(/*Id*/usize, /*Caller*/usize),
}

#[derive(Clone, Copy)]
pub struct TessellateConfig {
    pub origin:DVec3,
    pub scale:f64,
    pub max_depth:u16,
}
impl TessellateConfig {
    pub fn new() -> Self
    {
        Self {
            origin:DVec3::ZERO,
            scale:1.,
            max_depth:1,
        }
    }
}

#[derive(Clone, Copy)]
pub struct Vertex {
    pub position:DVec3,
    pub height:f64,
    pub color:Vec3,
}
impl Vertex {
    pub fn new() -> Self
    {
        Self {
            position:DVec3::ZERO,
            height:0.,
            color:Vec3::ONE,
        }
    }

    pub fn with(position:DVec3, height:f64) -> Self
    {
        Self {
            position:position,
            height:height,
            color:Vec3::ONE,
        }
    }
}

#[derive(Clone, Copy)]
pub struct Cell {
    pub depth:u16,
    pub direction:u8,
    pub sibling:u8,
    pub neighbors:[usize;3],
    pub vertices:[usize;3],
}
impl Cell {
    pub fn new() -> Self
    {
        Self {
            depth:0,
            direction:3,
            sibling:3,
            neighbors:[0; 3],
            vertices:[0; 3],
        }
    }
}

#[derive(Clone)]
pub struct Terrain {
    pub vertices:Pool<Vertex>,
    pub cells:Pool<Cell>,
    pub extent:[usize; 3],
    pub normal:DVec3,
    generator:Generator,

    pub config:TessellateConfig,
}
impl Terrain {
    pub fn new(generator:Generator) -> Self
    {
        Self {
            vertices:Pool::<Vertex>::new(),
            cells:Pool::<Cell>::new(),
            extent:[0; 3],
            normal:DVec3::Y,
            generator:generator,

            config:TessellateConfig::new(),
        }
    }

    pub fn triangle(basis:(DVec3, DVec3, DVec3), generator:Generator) -> Self
    {
        let mut grid = Self::new(generator);
        let (a, b, c) = basis;

        let cross = (c - a).cross(b - a);
        grid.normal = cross.normalize();

        grid.extent = [
            grid.add_vertex(a),
            grid.add_vertex(b),
            grid.add_vertex(c),
        ];

        grid.cells.add({
            let mut cell = Cell::new();
            cell.vertices = grid.extent;
            cell
        });

        grid
    }

    pub fn subdivide(&mut self)
    {
        for id in self.cells.list() {
            self.node_subdivide(id, 0);
        }
    }

    pub fn subdivide_cell(&mut self, id:usize)
    {
        self.node_update(&vec![Action::Subdivide(id, 0)]);
    }

    pub fn tessellate(&mut self, max_depth:u16)
    {
        let mut data = self.cells.list();
        let mut data_len = data.len();

        let mut config = self.config.clone();
        config.max_depth = max_depth;

        let mut passes = 1;
        while data.len() > 0 {
            let pass = data.clone();
            data.clear();

            for id in pass {
                if self.node_prepare(id, &config) {
                    data.push(id);
                }
            }

            let new_data = self.cells.list();
            if new_data.len() > data_len {
                data.append(&mut new_data[data_len..].to_vec());
                data_len = new_data.len();
            }

            if passes % 10 == 0 {
                println!("Pass {}", passes);
            }

            passes += 1;
        }
    }

    fn node_prepare(&mut self, id:usize, config:&TessellateConfig) -> bool
    // Test the height change of the node to determine if it should be subdivided.
    //
    {
        let cell = if let Some(data) = self.cells.get(id) {
            *data
        } else { return false; };

        // Determine if cell should be subdivided.
        let vt = [
            *self.vertices.get(cell.vertices[0]).unwrap(),
            *self.vertices.get(cell.vertices[1]).unwrap(),
            *self.vertices.get(cell.vertices[2]).unwrap(),
        ];
        let vp = [
            vt[0].position,
            vt[1].position,
            vt[2].position,
        ];
        let v = [
            vp[0] + (self.normal * vt[0].height),
            vp[1] + (self.normal * vt[1].height),
            vp[2] + (self.normal * vt[2].height),
        ];
        let normal = (v[1] - v[0]).cross(v[2] - v[0]).normalize();

        let mut delta: f64 = 0.;
        for n in 0..3 {
            let neighbor_id = cell.neighbors[n];
            if neighbor_id != 0 {
                if let Some(neighbor) = self.cells.get(neighbor_id) {
                    let vt = [
                        *self.vertices.get(neighbor.vertices[0]).unwrap(),
                        *self.vertices.get(neighbor.vertices[1]).unwrap(),
                        *self.vertices.get(neighbor.vertices[2]).unwrap(),
                    ];
                    let v = [
                        vt[0].position + (self.normal * vt[0].height),
                        vt[1].position + (self.normal * vt[1].height),
                        vt[2].position + (self.normal * vt[2].height),
                    ];

                    delta = delta.max(normal.angle_between((v[1] - v[0]).cross(v[2] - v[0]).normalize()));
                }
            }
        }

        let cvp = [
            vp[0] - config.origin,
            vp[1] - config.origin,
            vp[2] - config.origin,
        ];
        let centroid = (cvp[0] + cvp[1] + cvp[2]) / 3.;
        let min_dist = centroid.length_squared().min(cvp[0].length_squared()).min(cvp[1].length_squared()).min(cvp[2].length_squared()).sqrt() * config.scale;

        let depth_threshold = 0.125 * cell.depth as f64;
        let distance_threshold = min_dist.log2();

        //println!("dt {}", distance_threshold);

        let threshold_modifier = (depth_threshold + distance_threshold) * RADIAL_FALLOFF;
        let threshold: f64 = DELTA_THRESHOLD * threshold_modifier * threshold_modifier;
        if delta > threshold && delta > threshold_modifier && cell.depth < (config.max_depth * 2) {
            self.node_update(&vec![Action::Subdivide(id, 0)]);

            if let Some(data) = self.cells.get(id) {
                data.depth != cell.depth
            } else { false }
        } else { false }
    }

    fn node_update(&mut self, actions:&Vec<Action>)
    {
        let mut actions = actions.clone();
        while let Some(action) = actions.pop() {
            actions.append(&mut match action {
                Action::Check(id, caller) => self.node_check(id, caller),
                Action::Subdivide(id, caller) => self.node_subdivide(id, caller),
                Action::Split(id, caller) => {
                    self.node_split(id, caller);
                    vec![]
                },
            });
        }
    }

    fn node_check(&mut self, id:usize, caller:usize) -> Vec<Action>
    // Determine if node needs to be subdivided or split.
    //
    {
        let cell = if let Some(data) = self.cells.get(id) {
            *data
        } else { return vec![]; };

        //println!("caller {}", caller);
        self.print_cell("UPDATE", id, &cell);

        // Determine if relation with neighbors is stable.
        //  - Regular cell is stable if neighbors are same depth or depth +/- 1.
        //  - Split cell is stable if exactly 1 neighbor is depth +1 and rest are same or -1.
        // Count number of regular neighbors of greater depth.
        let mut diff_direction :usize = 0;
        let mut diff_count :usize = 0;
        for n in 0..cell.neighbors.len() {
            let neighbor_id = cell.neighbors[n];

            if neighbor_id != 0 {
                let neighbor = self.cells.get(neighbor_id).unwrap().clone();
                if neighbor.depth >= (cell.depth & !1) + 2 {
                    diff_count += 1;
                    diff_direction = n;
                }
            }
        }

        // Determine how cell should be handled.
        if cell.depth & 1 == 0 {
            match diff_count {
                0   => { vec![] }
                1   => { vec![Action::Split(id, diff_direction)] }
                2|3 => { vec![Action::Subdivide(id, caller)] }
                _   => { panic!("invalid diff count") }
            }
        }
        else {
            match diff_count {
                0   => { vec![] }
                1   => {
                    if cell.direction as usize != diff_direction {
                        vec![Action::Subdivide(id, caller)]
                    }
                    else {
                        vec![]
                    }
                }
                2|3 => { vec![Action::Subdivide(id, caller)] }
                _   => { panic!("invalid diff count") }
            }
        }

        // Determine if neighbors need to be updated.
    }

    fn node_subdivide(&mut self, id:usize, _caller:usize) -> Vec<Action>
    {
        let mut cell = if let Some(data) = self.cells.get(id) {
            *data
        } else { return vec![]; };

        let mut actions = Vec::<Action>::new();

        //println!("caller {}", caller);
        self.print_cell("BEGDIV", id, &cell);

        // If half-depth, get twin information.
        // If sibling 1, forward call to sibling 0.
        //
        let twin_dir = if cell.depth & 1 == 0 { 3 } else { Self::twin_neighbor(cell.direction, cell.sibling) };

        let mut twin_id = 0;
        let mut twin :Option<Cell> = None;

        let direction = cell.direction as usize;
        let d_offset_0 = (direction + 1) % 3;
        let d_offset_1 = (direction + 2) % 3;

        if cell.depth & 1 == 1 {
            twin_id = cell.neighbors[twin_dir];

            if cell.sibling == 1 {
                return vec![Action::Subdivide(twin_id, id)];
            }

            twin = Some(self.cells.get(twin_id).unwrap().clone());
        }

        // Get outer neighbors and vertices.
        let (neighbors, vertices) = match twin {
            Some(twin) => {
                let mut neighbors = cell.neighbors;
                let mut vertices = cell.vertices;

                neighbors[twin_dir] = twin.neighbors[twin_dir];
                vertices[d_offset_0] = twin.vertices[d_offset_0];

                (neighbors, vertices)
            }

            None => {
                (cell.neighbors, cell.vertices)
            }
        };

        let mut updates = Vec::<usize>::new();
        for n in 0..neighbors.len() {
            let neighbor_id = neighbors[n];
            if let Some(neighbor) = self.cells.get(neighbor_id) {
                if neighbor.neighbors.contains(&id) || (twin_id != 0 && neighbor.neighbors.contains(&twin_id)) {
                    updates.push(neighbor_id);
                }
                else {
                    let nd_offset_0 = (n + 2) % 3;
                    let nd_offset_1 = (n + 1) % 3;

                    updates.push(neighbor.neighbors[nd_offset_0]);
                    updates.push(neighbor.neighbors[nd_offset_1]);
                }
            }
        }

        if let Some(twin) = &twin {
            let neighbor_id = twin.neighbors[direction];
            if let Some(neighbor) = self.cells.get(neighbor_id) {
                if neighbor.neighbors[direction] == id || neighbor.neighbors[direction] == twin_id {
                    updates.push(neighbor_id);
                }
                else {
                    updates.push(neighbor.neighbors[d_offset_0]);
                    updates.push(neighbor.neighbors[d_offset_1]);
                }
            }
        }

        // If a neighbor is of lesser depth, that node needs to be subdivided
        // before subdividing the current node.
        // If node is subdivided as a result, no work to do.
        //
        let initial_depth = cell.depth;
        for n in 0..neighbors.len() {
            let neighbor_id = neighbors[n];
            if neighbor_id != 0 {
                let neighbor = *self.cells.get(neighbor_id).unwrap();
                if neighbor.depth < initial_depth && neighbor.depth & 1 == 1 {
                    actions.push(Action::Subdivide(neighbor_id, id));
                }

                // Refresh cell data incase changed.
                cell = if let Some(data) = self.cells.get(id) {
                    *data
                } else { return vec![]; };
            }
        }
        if !actions.is_empty() {
            return [vec![Action::Subdivide(id, id)], actions].concat();
        }


        // Get vector values of outer vertices.
        let vertex_values = [
            (*self.vertices.get(vertices[0]).unwrap()).position,
            (*self.vertices.get(vertices[1]).unwrap()).position,
            (*self.vertices.get(vertices[2]).unwrap()).position,
        ];


        // Get midpoints from adjacent nodes.
        let mut midpoints = [0; 3];
        for n in 0..neighbors.len() {
            let neighbor_id = neighbors[n];
            if neighbor_id != 0 {
                let neighbor = *self.cells.get(neighbor_id).unwrap();
                
                if (neighbor.depth > cell.depth && neighbor.depth & 1 == 0) || (neighbor.depth > cell.depth && neighbor.direction == n as u8) {
                    midpoints[n] = neighbor.vertices[n];
                }
            }
        }
        match twin {
            Some(_) => {
                midpoints[direction] = cell.vertices[d_offset_0];
            }
            None => { }
        }

        // Fill missing midpoint vertices.
        let subvertices = [
            if midpoints[0] != 0 { midpoints[0] } else { self.add_vertex(((vertex_values[2] - vertex_values[1]) / 2.) + vertex_values[1]) },
            if midpoints[1] != 0 { midpoints[1] } else { self.add_vertex(((vertex_values[0] - vertex_values[2]) / 2.) + vertex_values[2]) },
            if midpoints[2] != 0 { midpoints[2] } else { self.add_vertex(((vertex_values[1] - vertex_values[0]) / 2.) + vertex_values[0]) },
        ];


        // Prepare child cells.
        let mut children = [
            Cell::new(),
            Cell::new(),
            Cell::new(),
            Cell::new(),
        ];
        let child_ids = [
            self.cells.add(Cell::new()),
            self.cells.add(Cell::new()),
            if twin.is_some() { twin_id } else { self.cells.add(Cell::new()) },
            id,
        ];

        cell.depth = (cell.depth + 2) & !1;
        for c in 0..children.len() {
            children[c].depth = cell.depth;
            children[c].sibling = c as u8;
        }
        for c in 0..3 {
            children[3].neighbors[c] = child_ids[c];
            children[c].neighbors[c] = id;
        }

        // Set new cell vertices.
        children[0].vertices = [    vertices[0], subvertices[2], subvertices[1] ];
        children[1].vertices = [ subvertices[2],    vertices[1], subvertices[0] ];
        children[2].vertices = [ subvertices[1], subvertices[0],    vertices[2] ];
        children[3].vertices = [ subvertices[0], subvertices[1], subvertices[2] ];


        // Link twin neighbor to center cell.
        if twin.is_some() {
            if let Some(neighbor) = self.cells.get_mut(neighbors[twin_dir]) {
                if neighbor.neighbors[twin_dir] == twin_id {
                    neighbor.neighbors[twin_dir] = id;
                }
            }
        }

        // Trade neighbor links with neighboring cells.
        for n in 0..neighbors.len() {
            let neighbor_id = neighbors[n];

            let nd_offset_0 = (n + 2) % 3;
            let nd_offset_1 = (n + 1) % 3;

            //println!("neighbor {} {}", n, neighbor_id);

            let mut processed = false;
            if n == direction {
                if let Some(twin) = &twin {
                    // Update both children that have edges in direction.

                    let n0 = cell.neighbors[direction];
                    let n1 = twin.neighbors[direction];

                    children[nd_offset_0].neighbors[direction] = n0;
                    children[nd_offset_1].neighbors[direction] = n1;

                    //println!("d {} [{}] = {}, {} [{}] = {}", n0, n, child_ids[nd_offset_0], n1, n, child_ids[nd_offset_1]);

                    if let Some(neighbor) = self.cells.get_mut(n0) {
                        neighbor.neighbors[n] = child_ids[nd_offset_0];
                    }
                    if let Some(neighbor) = self.cells.get_mut(n1) {
                        neighbor.neighbors[n] = child_ids[nd_offset_1];
                    }

                    processed = true;
                }
            }
            if !processed {
                if neighbor_id != 0 {
                    let neighbor = self.cells.get(neighbor_id).unwrap().clone();

                    if neighbor.depth == cell.depth {
                        // Update both children that have edges in direction.
                        // Neighbor must be parent.

                        let n0 = neighbor.neighbors[nd_offset_1];
                        let n1 = neighbor.neighbors[nd_offset_0];

                        //println!("{} [{}] = {}, {} [{}] = {}", n0, n, child_ids[nd_offset_0], n1, n, child_ids[nd_offset_1]);

                        children[nd_offset_0].neighbors[n] = n0;
                        children[nd_offset_1].neighbors[n] = n1;

                        if let Some(neighbor) = self.cells.get_mut(n0) {
                            neighbor.neighbors[n] = child_ids[nd_offset_0];
                        }
                        if let Some(neighbor) = self.cells.get_mut(n1) {
                            neighbor.neighbors[n] = child_ids[nd_offset_1];
                        }
                    }
                }
            }
        }

        // Update cells.
        for i in 0..children.len() {
            self.print_cell("SUBDIV", child_ids[i], &children[i]);

            self.cells.update(child_ids[i], children[i]).ok();
        }

        // Update neighbor cells
        for update in updates {
            if update != 0 {
                actions.push(Action::Check(update, id));
            }
        }
        actions
    }

    fn node_split(&mut self, id:usize, direction:usize)
    {
        let mut cell = if let Some(data) = self.cells.get(id) {
            *data
        } else { return; };
        if cell.direction as usize == direction { return; }

        self.print_cell("SPLIT", id, &cell);

        let neighbors = cell.neighbors;

        if cell.depth & 1 == 0 {
            let mut twin = Cell::new();

            // Set cell and twin properties.
            cell.depth += 1;
            twin.depth = cell.depth;

            cell.direction = direction as u8;
            twin.direction = direction as u8;

            cell.sibling = 0;
            twin.sibling = 1;

            let neighbor_id = cell.neighbors[direction];
            let neighbor = self.cells.get(neighbor_id).unwrap().clone();
            let midpoint = neighbor.vertices[direction];

            let d_offset_0 = (direction + 1) % 3;
            let d_offset_1 = (direction + 2) % 3;

            // Set cell and twin vertices.
            twin.vertices = cell.vertices;
            twin.vertices[d_offset_1] = midpoint;
            cell.vertices[d_offset_0] = midpoint;

            // Set cell and twin neighbors.
            let n0 = neighbor.neighbors[d_offset_0];
            let n1 = neighbor.neighbors[d_offset_1];

            twin.neighbors = cell.neighbors;
            twin.neighbors[direction] = n1;
            twin.neighbors[Self::twin_neighbor(direction as u8, 1)] = id;

            let twin_id = self.cells.add(twin);
            cell.neighbors[direction] = n0;
            cell.neighbors[Self::twin_neighbor(direction as u8, 0)] = twin_id;

            // Update links for outer neighbors
            if let Some(neighbor) = self.cells.get_mut(n0) {
                neighbor.neighbors[direction] = id;
            }
            if let Some(neighbor) = self.cells.get_mut(n1) {
                neighbor.neighbors[direction] = twin_id;
            }
            if let Some(neighbor) = self.cells.get_mut(neighbors[d_offset_1]) {
                neighbor.neighbors[d_offset_1] = twin_id;
            }

            self.print_cell("SUBHAV", id, &cell);
            self.print_cell("SUBHAV", twin_id, &twin);

            // Push updated data
            self.cells.update(id, cell).ok();
        }

        // Node is already split, so cells only need to be rotated.
        // This probably cannot happen, as it would require a neighbor to disappear
        //  and another to form before updating of this triangle.
        else {
            panic!("split node needs to be rotated");
        }
    }

    fn twin_neighbor(direction:u8, sibling:u8) -> usize
    {
        ((direction + 2 - sibling) % 3) as usize
    }

    fn add_vertex(&mut self, vertex:DVec3) -> usize
    {
        self.vertices.add(Vertex::with(vertex, self.generator.generate(vertex)))
    }

    fn print_cell(&self, text:&str, id:usize, cell:&Cell)
    {
        if DEBUG_PRINTS {
            println!("{:8} {:3} ( {:2} ) [{:3} {:3} {:3}] [{:3} {:3} {:3}]",
                text,
                id,
                cell.depth,
                cell.vertices[0],
                cell.vertices[1],
                cell.vertices[2],
                cell.neighbors[0],
                cell.neighbors[1],
                cell.neighbors[2],
            );
        }
    }

    pub fn to_mesh(&self) -> crate::mesh::Mesh
    {
        let mut mesh = crate::mesh::Mesh::new();
        let mut conversion = Sparse::<usize>::new();

        // Add vertices
        for index in self.vertices.list() {
            if let Some(vertex) = self.vertices.get(index) {
                conversion.set(index as isize, mesh.vertices.len());
                mesh.vertices.push(crate::mesh::Vertex::new(
                    (vertex.position + (self.normal * vertex.height)).as_vec3(),
                    vertex.color,
                ));
            }
        }

        // Add faces
        for index in self.cells.list() {
            if let Some(cell) = self.cells.get(index) {
                self.print_cell("FINAL", index, &cell);

                let a = *conversion.get(cell.vertices[0] as isize).unwrap() as i32;
                let b = *conversion.get(cell.vertices[1] as isize).unwrap() as i32;
                let c = *conversion.get(cell.vertices[2] as isize).unwrap() as i32;
                
                mesh.faces.push(IVec3::new(a, c, b));
            }
        }

        mesh
    }
}
